package provide SyncTimer 1.1

##  synctimer.tcl
##
##  SyncTimer - timers synchronized with clock
##
##  This package provides commands able to start/stop synchronized timers.
##  Each time a timer is triggered, it executes a user-defined command.
##  Timers are triggered at regular intervals, and are aligned (synchronized)
##  with the system-clock.
##  i.e. timers can be triggered when the system-clock strikes the hours
##  ( or the half-hours, or more ... read manual)
##
##  Copyright (c) 2005 <Irrational Numbers> : <aburatti@libero.it>
##
##  This program is free software; you can redistibute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation.
##

namespace eval SyncTimer {
    namespace export every cancel clockTuning
    
    variable ID 0
    variable SYNC_INFO ; # array
    
     # It has been observed that the system-clock strikes the time (seconds)
     #  after N milliseconds, compared to the tick-clock [clock clicks -milliseconds] 
     # By experiments, I've found N is equal to 40 milliseconds.
     #  ( i.e. when the tick-clock is between  XXXX039 and XXXX041 milliseconds,
     #     the system-clock is updated) 
     # I don't know why ... but it works.
     # Since this value is stored in the public variable 'clockTuning',
     #  you can set a different value in your application code.    
    variable clockTuning 40 ; # experimentally found
    
    proc every {deltaStr basetimeStr args} {
        variable ID
        incr ID
        set delta [clock scan $deltaStr -base 0]
        set basetime [clock scan $basetimeStr]
        eval _sync $ID $delta $basetime $args
        return $ID
    }
    
    proc _sync {id delta basetime args} {
        variable SYNC_INFO
        variable clockTuning

        if { $delta == 0 } {
            set dt 0
        } else {
            set now [clock seconds]
            set msec [expr [clock clicks -milliseconds]%1000]
            set dt [expr ($delta-($now-$basetime)%$delta)*1000-$msec+$clockTuning]       
        }
        if { $dt == 0 } {
            # run once
            set SYNC_INFO($id) [eval after $dt $args]
        } else {
            # start after $dt, and then reschedule
            set SYNC_INFO($id) [eval after $dt $args [list \; [namespace code _sync] $id $delta $basetime $args ]]
        }
                
    }
    
    proc cancel {id} {
        variable SYNC_INFO
        after cancel $SYNC_INFO($id)
        unset SYNC_INFO($id)
    }
}

